"""
Systems package of CoSApp.
"""
from cosapp.systems.system import System, ConversionType
from cosapp.systems.systemfamily import SystemFamily
from cosapp.systems.metamodels import MetaSystem
from cosapp.systems.externalsystem import TCPSystem
from cosapp.systems.processsystem import ProcessSystem

from cosapp.utils.surrogate_models import (
    FloatKrigingSurrogate,
    LinearNearestNeighbor,
    WeightedNearestNeighbor,
    RBFNearestNeighbor,
    ResponseSurface,
)

__all__ = [
    "System",
    "SystemFamily",
    "MetaSystem",
    "LinearNearestNeighbor",
    "RBFNearestNeighbor",
    "WeightedNearestNeighbor",
    "FloatKrigingSurrogate",
    "ResponseSurface",
    "ConversionType",
    "ProcessSystem",
    "TCPSystem",
]



import pytest
import numpy as np

from cosapp.drivers.time.scenario import TimeAssignString
from .conftest import ScalarOde, VectorOde


def test_TimeAssignString__init__():
    ode = ScalarOde('ode')
    interpolator = lambda t: t if t < 1 else 3 - 2 * t
    bc = TimeAssignString('df', interpolator, ode)
    assert bc.rhs is interpolator

    with pytest.raises(TypeError, match="must be a callable function"):
        TimeAssignString('df', 0.5, ode)


def test_TimeAssignString_exec_scalar(clock):
    ode = ScalarOde('ode')
    interpolator = lambda t: t if t < 1 else 3 - 2 * t
    bc = TimeAssignString('df', interpolator, ode)

    clock.time = 999

    for t in (0, 0.8, 1, 2, 3.4):
        bc.exec(t)
        assert ode.df == interpolator(t)
        assert ode.time == 999


def test_TimeAssignString_exec_vector(clock):
    ode = VectorOde('ode')
    interpolator = lambda t: np.array([t, np.exp(-t), np.sin(t)])
    bc = TimeAssignString('dv', interpolator, ode)
    assert bc.rhs is interpolator

    clock.time = 999

    for t in (0, 0.8, 1, 2, 3.4, -2.5):
        bc.exec(t)
        assert np.array_equal(ode.dv, interpolator(t))
        assert ode.time == 999


def test_TimeAssignString_exec_masked_array(clock):
    ode = VectorOde('ode')
    ode.dv[:] = -0.99
    interpolator = lambda t: np.array([t, np.exp(-t)])
    bc = TimeAssignString('dv[::2]', interpolator, ode)
    assert bc.rhs is interpolator

    clock.time = 999

    for t in (0, 0.8, 1, 2, 3.4, -2.5):
        bc.exec(t)
        assert np.array_equal(ode.dv[::2], interpolator(t))
        assert ode.dv[1] == -0.99
        assert ode.time == 999



"""
Configuration of CoSApp.
"""
import os
import json

try:
    from json import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError
import logging
from pathlib import Path
from typing import FrozenSet, Union

import jsonschema

logger = logging.getLogger(__name__)


class CoSAppConfiguration:
    r"""Encapsulate CoSApp configuration parameters and its handlers.
    
    By default, CoSApp configuration folder is stored in 
    - `$HOME/.cosapp.d` (Linux)
    - `%USERPROFILE%\.cosapp.d` (MS Windows)

    This default configuration folder is overwritten by the environment
    variable ``COSAPP_CONFIG_DIR``.
    """

    COSAPP_CONFIG_DIR = ".cosapp.d"
    CONFIG_FILE = "cosapp_config.json"

    def __init__(self) -> None:
        """Constructor"""
        self._userid = ""
        self._roles: FrozenSet[FrozenSet[str]] = frozenset()

        self.__load_configuration()

    def __load_configuration(self) -> None:
        """Read configuration from file or generate the default.

        Raises
        ------
        OSError
            If the current platform is not recognized
        """
        fullpath = self.get_config_filename()
        ok = False

        if os.path.isfile(fullpath):  # Load local parameters - offline connection
            try:
                parameters = self.validate_file(fullpath)
            except (OSError, JSONDecodeError):
                ok = False
            else:
                ok = True
                self._userid = parameters["userid"]
                self._roles = frozenset(map(frozenset, parameters["roles"]))

        if not ok:
            logger.warn(
                f"Configuration file `{fullpath!s}` cannot be opened; fall back to default."
            )
            try:
                self.update_userid()
            except OSError:
                self._userid = unknwon_id = "UNKNOWN"
                self._roles = frozenset()
                logger.warn(
                    f"Unable to determine user ID; fall back to {unknwon_id!r}."
                )
            # Try to update user permission from official root
            self.update_configuration()

    @staticmethod
    def get_config_dir() -> Path:
        try:
            return Path(os.environ["COSAPP_CONFIG_DIR"])
        except KeyError:
            return Path.home().joinpath(CoSAppConfiguration.COSAPP_CONFIG_DIR)

    @classmethod
    def get_config_filename(cls) -> Path:
        config_dir = cls.get_config_dir()
        return config_dir.joinpath(cls.CONFIG_FILE)

    @property
    def userid(self) -> str:
        """str : User ID"""
        return self._userid

    @property
    def roles(self) -> FrozenSet[FrozenSet[str]]:
        """FrozenSet[FrozenSet[str]] : Roles assigned to user."""
        return self._roles

    @staticmethod
    def config_schema() -> dict:
        """Static method returning the JSON validation schema of the class."""
        path = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(path , "configuration_schema.json"), "r") as fp:
            config_schema = json.load(fp)
        return config_schema

    @classmethod
    def validate_file(cls, filename: Union[str, Path]) -> dict:
        """Validate the provided file against JSON schema for configuration file.

        Parameters
        ----------
        filename : str or Path
            Absolute path to the file to be tested.

        Returns
        -------
        dict
            The dictionary read in the validated file

        Raises
        ------
        `OSError` (and derived exceptions)
            If a problem occurs while opening the file.
        `jsonschema.exceptions.ValidationError`
            If the provided file does not conform to the JSON schema.
        """
        with open(filename, "r") as fp:
            params = json.load(fp)

        jsonschema.validate(params, cls.config_schema())

        return params

    def update_userid(self) -> None:
        """Update current user id.
        """
        candidates = iter(["USERNAME", "USER"])
        old_id = self._userid

        while not self._userid:
            try:
                self._userid = os.environ.get(next(candidates), "")
            except StopIteration:
                raise OSError("Unable to determine user ID.")
        
        if self._userid != old_id:
            logger.info(f"User ID changed from {old_id!r} to {self._userid!r}.")

    def update_configuration(self) -> None:
        """Update current user configuration file.
        """
        # TODO - request reference roles source here

        # Save back changes following the update
        parameters = {
            "userid": self.userid,
            "roles": list(map(list, self.roles)),
        }

        try:
            config_file = self.get_config_filename()
            config_dir = os.path.dirname(config_file)
            if not os.path.isdir(config_dir):
                os.makedirs(config_dir)

            with open(config_file, "w") as file:
                json.dump(parameters, file, sort_keys=True)

        except OSError:
            logger.warning("Failed to save configuration locally.")



import numpy, scipy.interpolate
import enum
import logging
import warnings
from typing import Any, Dict, List, Tuple, Callable

from cosapp.systems import System
from cosapp.drivers.driver import Driver
from cosapp.multimode.event import Event
from cosapp.core.numerics.boundary import Boundary
from cosapp.core.eval_str import AssignString
from cosapp.utils.naming import natural_varname
from cosapp.utils.helpers import check_arg

logger = logging.getLogger(__name__)


class Interpolator:
    """Class describing a function defined from tabulated data,
    based on interpolating function `interpolator`. Function `interpolator`
    must return a callable function x -> f(x), interpolating [y_0, ..., y_n] at
    points [x_0, ..., x_n]."""

    class Kind(enum.Enum):
        """Interpolator kind"""
        Linear = scipy.interpolate.interp1d
        CubicSpline = scipy.interpolate.CubicSpline
        Pchip = scipy.interpolate.PchipInterpolator
    
    def __init__(self, data, kind=Kind.Linear):
        self.__data = None
        self.kind = kind
        self.data = data

    @property
    def kind(self) -> Kind:
        return self.__kind

    @kind.setter
    def kind(self, kind: Kind) -> None:
        check_arg(kind, 'kind', self.Kind)
        self.__kind = kind
        self.__update_evaluator()

    @property
    def data(self) -> numpy.ndarray:
        return self.__data
    
    @data.setter
    def data(self, data) -> None:
        data = numpy.asarray(data, dtype=float)
        shape = data.shape
        if data.ndim != 2:
            raise ValueError(f"invalid shape {shape}; `data` must be a 2D array.")
        if shape[1] == 2:
            # data given as [[x0, y0], ..., [xn, yn]]
            data = data.T
        if data.shape[1] < 2:
            raise ValueError(f"data must contain at least two points")
        # sort data to ensure increasing x
        self.__data = data[:, data[0].argsort()]
        self.__data.setflags(write=0)
        self.__update_evaluator()

    def __update_evaluator(self) -> None:
        data = self.data
        if data is None:
            def evaluator(x):
                raise ValueError("data was not specified")
        else:
            evaluator = self.kind.value(data[0], data[1])
        self.__evaluator = evaluator

    def __call__(self, t: float) -> numpy.ndarray:
        return self.__evaluator(t)


class TimeAssignString:
    """Creates an executable assignment to handle time boundary conditions
    of the kind `lhs = F(t, data)`, where F is a function of some dataset at time t,
    and where `lhs` refers to a variable name in system `context`.
    This class is very similar to `cosapp.core.eval_str.AssignString`,
    except the right-hand side is a callable function.
    """
    def __init__(self, lhs: str, rhs: Callable[[float], Any], context: System):
        if not callable(rhs):
            raise TypeError(
                f"right-hand side must be a callable function; got {rhs!r}"
            )
        Boundary.parse(context, lhs, inputs_only=True)  # checks that variable is valid
        fname = f"BC{id(rhs)}"
        assignment = f"{context.name}.{lhs} = {fname}(t)"
        self.__locals = {fname: rhs, context.name: context, 't': 0}
        self.__code = compile(assignment, "<string>", "single")  # type: CodeType
        self.__str = f"{lhs} = {type(rhs).__name__}(t)"
        self.__rhs = rhs

    def exec(self, t: float) -> None:
        """Evaluates rhs(t), and executes assignment lhs <- rhs(t).
        """
        self.__locals['t'] = t
        exec(self.__code, {}, self.__locals)

    def __str__(self) -> str:
        return self.__str

    @property
    def rhs(self) -> Callable[[float], Any]:
        """Callable: assignment right-hand side function"""
        return self.__rhs
    
    @property
    def constant(self):
        return False


class InterpolAssignString(TimeAssignString):
    """Creates an executable assignment to handle time boundary conditions
    of the kind `lhs = F(t, data)`, where F is an interpolation function of some dataset at time t,
    and where `lhs` refers to a variable name in system `context`.
    This class is very similar to `cosapp.core.eval_str.AssignString`, except the right-hand side is a function.
    In order to limit the scope of the callable function, the rhs can only be of type `Interpolator`.
    """
    def __init__(self, lhs: str, rhs: Callable[[float], Any], context: System):
        # Strict type check (as opposed to `isinstance`), to ensure the type of
        # `function` is `Interpolator`, and is not derived from `Interpolator`.
        if type(rhs) is not Interpolator:
            raise TypeError(
                f"Functions used in time boundary conditions may only be of type `Interpolator`"
                f"; got {type(rhs)}"
            )
        super().__init__(lhs, rhs, context)
        self.__context = context

    def exec(self) -> None:
        """Evaluates rhs at context time, and executes assignment lhs <- rhs."""
        super().exec(self.__context.time)


class Scenario:
    """Class managing boundary and initial conditions for time simulations"""

    def __init__(self, name: str, owner: Driver) -> None:
        """Initialize object

        Parameters
        ----------
        name: str
            Name of the `Module`
        owner : Driver
            :py:class:`~cosapp.drivers.driver.Driver` to which object belong
        """
        self.__case_values = []   # type: List[AssignString]
        self.__init_values = []   # type: List[AssignString]
        self.__stop: Event = None
        self.name = name
        self.owner = owner

    @classmethod
    def make(cls, name: str, driver: Driver, init: Dict[str, Any], values: Dict[str, Any]) -> "Scenario":
        """Scenario factory"""
        scenario = cls(name, driver)
        scenario.set_init(init)
        scenario.set_values(values)
        return scenario

    def apply_init_values(self) -> None:
        """Execute assignments corresponding to initial conditions"""
        logger.debug("Apply initial conditions")
        for assignment in self.__init_values:
            logger.debug(f"\t{assignment}")
            assignment.exec()

    def update_values(self) -> None:
        """Execute assignments corresponding to boundary conditions"""
        for assignment in self.__case_values:
            assignment.exec()

    @property
    def context(self) -> System:
        """System: evaluation context of initial and boundary conditions,
        that is the system controled by owner driver"""
        return self.__context

    @property
    def owner(self) -> Driver:
        """Driver: owner driver"""
        return self.__owner

    @owner.setter
    def owner(self, driver: Driver) -> None:
        check_arg(driver, "owner", Driver)
        self.__owner = driver
        self.__context = context = driver.owner  # type: System
        if context is None:
            self.__stop = None
        else:
            self.__stop = Event('stop', context, desc="Stop criterion", final=True)
        self.clear_init()
        self.clear_values()

    @property
    def stop(self) -> Event:
        """Event: discrete event triggering the end of scenario"""
        return self.__stop

    def add_init(self, modifications: Dict[str, Any]) -> None:
        """Add a set of initial conditions, from a dictionary of the kind {'variable': value, ...}

        Parameters
        ----------
        modifications : Dict[str, Any]
            Dictionary of (variable name, value)

        Examples
        --------
        >>> scenario.add_init({'myvar': 42, 'port.dummy': '-2 * alpha'})
        """
        check_arg(modifications, 'modifications', dict,
            lambda d: all(isinstance(key, str) for key in d.keys())
        )
        if self.owner is None:
            raise AttributeError(f"Driver {self.name!r} must be attached to a System to set initial values.")

        for varname, value in modifications.items():
            varname, context = self._get_alias(varname, inputs_only=False)
            if context is None:
                continue
            assignment = AssignString(varname, value, context)
            if assignment.constant:
                # If assignment is constant, it is safer to insert it at the top
                # of the list, since other initial condition assignments might use it.
                self.__init_values.insert(0, assignment)
            else:
                self.__init_values.append(assignment)

    def set_init(self, modifications: Dict[str, Any]) -> None:
        """Set initial conditions, from a dictionary of the kind {'variable': value, ...}

        See `add_init` for further detail.
        """
        self.clear_init()
        self.add_init(modifications)

    def add_values(self, modifications: Dict[str, Any]) -> None:
        """Add a set of variables to the list of case values, from a dictionary of the kind {'variable': value, ...}

        Each variable and its value can be contextual, as in {'child1.port2.var': '2 * child2.foo'},
        as long as they are both evaluable in the context of the driver's owner.
        Explicit time dependency can be given using variable 't' in values, as in 'exp(-t / tau)'

        Parameters
        ----------
        modifications : Dict[str, Any]
            Dictionary of (variable name, value)

        Examples
        --------
        >>> scenario.add_values({'myvar': 42, 'port.dummy': 'cos(omega * t)'})
        """
        check_arg(modifications, 'modifications', dict,
            lambda d: all(isinstance(key, str) for key in d.keys())
        )
        if self.owner is None:
            raise AttributeError(f"Driver {self.name!r} must be attached to a System to set case values.")

        for varname, value in modifications.items():
            varname, context = self._get_alias(varname, inputs_only=True)
            if context is None:
                continue
            if callable(value):
                assignment = InterpolAssignString(varname, value, context)
            else:
                assignment = AssignString(varname, value, context)
            if assignment.constant:
                # If assignment is constant, it can be regarded as an initial condition,
                # rather than a time-dependent boundary condition.
                # Moreover, it is safer to insert it at the top of the list,
                # since other initial condition assignments might use it.
                self.__init_values.insert(0, assignment)
            else:
                self.__case_values.append(assignment)

    def set_values(self, modifications: Dict[str, Any]) -> None:
        """Set case values, from a dictionary of the kind {'variable': value, ...}

        See `add_values` for further detail.
        """
        self.clear_values()
        self.add_values(modifications)

    def clear_values(self) -> None:
        """Clears the list of boundary conditions"""
        self.__case_values.clear()

    def clear_init(self) -> None:
        """Clears the list of initial conditions"""
        self.__init_values.clear()

    @property
    def case_values(self) -> List[AssignString]:
        """List[AssignString]: list of boundary conditions"""
        return self.__case_values

    @property
    def init_values(self) -> List[AssignString]:
        """List[AssignString]: list of initial conditions"""
        return self.__init_values

    def _get_alias(self, lhs: str, inputs_only=True) -> Tuple[str, System]:
        """Resolve potential aliasing for variable `lhs`, targetted
        in an initial or a boundary condition.

        Arguments:
        ----------
        - lhs [str]:
            Assignment left-hand-side, i.e. targetted variable
        - inputs_only [bool, optional]:
            If `True` (default), only input variables are regarded as valid

        Returns:
        --------
        (free_lhs, context) [Tuple[str, System]]:
            Free lhs and its evaluation context, usable in `AssignString`.
        """
        context = self.__context
        info = Boundary.parse(context, lhs, inputs_only=inputs_only)  # checks that variable is valid

        if info.port.is_output:
            return (lhs, context)

        varname = natural_varname(info.basename)
        variable = info.ref
        try:
            alias = context.input_mapping[varname]
        except KeyError:
            warnings.warn(
                f"Skip connected variable {varname!r} in time scenario."
            )
            return None, None

        aliased = (variable is not alias)

        if not aliased:
            return (lhs, context)
        
        # Resolve aliasing
        port = alias.mapping
        alias_name = natural_varname(f"{port.name}.{alias.key}")
        try:
            path = context.get_path_to_child(port.owner)
        except ValueError:
            fullname = f"{port.owner.full_name()}.{alias_name}"
            warnings.warn(
                f"Variable {varname!r} is aliased by {fullname!r}"
                f", defined outside the context of {context.name!r}"
                f"; it is likely to be overwritten after the computation."
            )
            path = None
            alias_name = varname
            eval_context = context
        else:
            eval_context = alias.context
            if path:
                alias_name = f"{path}.{alias_name}"
            logger.info(
                f"Replace {varname!r} by {alias_name!r} in time scenario."
            )
        lhs = lhs.replace(varname, alias_name)  # capture mask, if any
        return (lhs, eval_context)

    def __repr__(self) -> str:
        s = f"{type(self).__name__} {self.name!r}, in {self.context.name!r}"
        def conditions(assigments, title):
            return "\n  - ".join(
                [f"\n{title.title()}:"] + list(map(str, assigments))
             ) if assigments else ""
        s += conditions(self.init_values, "Initial values")
        s += conditions(self.case_values, "Boundary conditions")
        return s



import pytest
import numpy as np

from cosapp.drivers.time.utils import TwoPointCubicInterpolator


@pytest.mark.parametrize("xs", [[1, 2], [0, 0.5], [-0.2, 0.1]])
@pytest.mark.parametrize("ys, dy", [
    ([0.7, -0.2], [-0.5, 0.1]),
    ([0, 1], [0, 1]),
    (   # 3D vector data
        [[-0.7, 0.1, 0.4], [1, -0.3, 0.2]],
        [[0.22, 1.4, 0.0], [0, 0.9, -0.6]],
    ),
])
def test_TwoPointCubicInterpolator(xs, ys, dy):
    xs = np.asarray(xs)
    ys = np.asarray(ys)
    dy = np.asarray(dy)
    f = TwoPointCubicInterpolator(xs, ys, dy)

    assert f(xs) == pytest.approx(np.transpose(ys), abs=1e-14)



import pytest
import logging
import re
import weakref
import itertools
from typing import MappingView
from unittest import mock
from contextlib import nullcontext as does_not_raise

from cosapp.core.module import Module
from cosapp.core.signal import Slot
from cosapp.drivers import Driver
from cosapp.utils.logging import LogFormat, LogLevel
from cosapp.utils.testing import no_exception


@pytest.fixture(autouse=True)
def PatchModule():
    """Patch Module to make it instanciable for tests"""
    patcher = mock.patch.multiple(
        Module,
        __abstractmethods__ = set(),
        is_standalone = lambda self: False,
    )
    patcher.start()
    yield
    patcher.stop()


@pytest.fixture(autouse=True)
def Russian_dolls():
    def factory(name: str, *submodules: str):
        head = Module(name)
        module_list = [head]
        last = head
        for name in submodules:
            last = last.add_child(Module(name))
            module_list.append(last)
        return module_list

    return factory


@pytest.fixture(scope="function")
def fake():
    return Module("fake")


@pytest.fixture
def composite():
    """Generates test module tree:

                 a
         ________|________
        |        |        |
       aa       ab        ac
      __|__           ____|____
     |     |         |    |    |
    aaa   aab       aca  acb  acc
    """
    def add_children(module, names):
        prefix = module.name
        for name in names:
            module.add_child(Module(f"{prefix}{name}"))

    a = Module('a')
    add_children(a, list('abc'))
    add_children(a.aa, list('ab'))
    add_children(a.ac, list('abc'))

    return a


@pytest.mark.parametrize("name, error", [
    ("a", None),
    ("A", None),
    ("foobar", None),
    ("foo4bar", None),
    ("loveYou2", None),
    ("CamelBack", None),
    ("foo_bar", None),
    ("foobar_", None),
    ("_foobar", ValueError),
    ("foo bar", ValueError),
    ("foobar?", ValueError),
    ("foo.bar", ValueError),
    ("foo:bar", ValueError),
    ("foo/bar", ValueError),
    ("1foobar", ValueError),
    ("foobar-2", ValueError),
    ("foobar:2", ValueError),
    ("foobar.2", ValueError),
    ("foo.bar", ValueError),
    ("inwards", ValueError),
    ("outwards", ValueError),
    (23, TypeError),
    (1.0, TypeError),
    (dict(a=True), TypeError),
    (list(), TypeError),
])
def test_Module__init__(name, error):
    """Test Module initialization"""
    if error is None:
        module = Module(name)
        assert module.name == name
        assert module.description == ""

    else:
        with pytest.raises(error):
            Module(name)


@pytest.mark.parametrize(
    "desc, expected", [
        ("A really great module", does_not_raise()),
        ("~~ non-alphanumeric start ~~", does_not_raise()),
        ("", does_not_raise()),
        (None, pytest.raises(TypeError)),
        (0, pytest.raises(TypeError)),
        (0.0, pytest.raises(TypeError)),
        (list(), pytest.raises(TypeError)),
        (dict(cool=True), pytest.raises(TypeError)),
    ],
)
def test_Module_description(desc: str, expected):
    """Test `description` getter & setter"""
    module = Module('foo')
    assert module.name == "foo"
    assert module.description == ""

    with expected:
        module.description = desc
        assert module.description == desc


def test_Module__weakref__(fake):
    with no_exception():
        proxy = weakref.proxy(fake)


def test_Module_name_setter():
    module = Module("foo")
    assert module.name == "foo"
    module.name = "bar"
    assert module.name == "bar"
    with pytest.raises(ValueError):
        module.name = "invalid.name"


def test_Module_contextual_name(Russian_dolls):
    a, b, c, d = Russian_dolls(*iter('abcd'))

    assert a.contextual_name == ""
    assert b.contextual_name == "b"
    assert c.contextual_name == "b.c"
    assert d.contextual_name == "b.c.d"


def test_Module_fullname(Russian_dolls):
    a, b, c, d = Russian_dolls(*iter('abcd'))

    assert a.full_name(trim_root=True) == ""
    assert b.full_name(trim_root=True) == "b"
    assert c.full_name(trim_root=True) == "b.c"
    assert d.full_name(trim_root=True) == "b.c.d"

    assert a.full_name() == "a"
    assert b.full_name() == "a.b"
    assert c.full_name() == "a.b.c"
    assert d.full_name() == "a.b.c.d"


def test_Module_path(Russian_dolls):
    a, b, c, d = Russian_dolls(*iter('abcd'))

    assert a.path() == [a]
    assert b.path() == [a, b]
    assert c.path() == [a, b, c]
    assert d.path() == [a, b, c, d]


def test_Module_root(Russian_dolls):
    a, b, c, d = Russian_dolls(*iter('abcd'))

    assert a.root() is a
    assert b.root() is a
    assert c.root() is a
    assert d.root() is a


def test_Module_path_namelist(Russian_dolls):
    a, b, c, d = Russian_dolls(*iter('abcd'))

    assert a.path_namelist() == ["a"]
    assert b.path_namelist() == ["a", "b"]
    assert c.path_namelist() == ["a", "b", "c"]
    assert d.path_namelist() == ["a", "b", "c", "d"]


def test_Module_compute_calls():
    m = Module("m")

    assert m.compute_calls == 0

    m.run_once()
    assert m.compute_calls == 1

    m.call_setup_run()  # Reset counter
    assert m.compute_calls == 0


def test_Module_add_child_desc():
    top = Module("top")
    foo = top.add_child(Module("foo"))
    bar = top.add_child(Module("bar"), desc="A great sub-module")
    assert foo.description == ""
    assert bar.description == "A great sub-module"


def test_Module_size():
    s = Module("s")
    T = Module("T")
    u = Module("u")
    v = Module("v")

    s.add_child(T)
    T.add_child(u)
    T.add_child(v)
    s.add_child(Module("r"))

    assert s.size == 5
    assert u.size == 1
    assert T.size == 3


def test_Module_get_path_to_child(Russian_dolls):
    a, b, c, d = Russian_dolls("a", "b", "c", "d")

    assert a.get_path_to_child(a) == ""
    assert a.get_path_to_child(b) == "b"
    assert a.get_path_to_child(c) == "b.c"
    assert a.get_path_to_child(d) == "b.c.d"
    assert b.get_path_to_child(c) == "c"
    assert b.get_path_to_child(d) == "c.d"

    with pytest.raises(ValueError, match="not a child of 'b'"):
        b.get_path_to_child(a)

    orphan = Module('orphan')
    with pytest.raises(ValueError, match="not a child of 'a'"):
        a.get_path_to_child(orphan)


def test_Module_pop_child_driver():
    # From Driver
    d = Driver("dummy")
    assert d.children == {}

    update = d.add_child(Driver("mydriver"))
    assert d.children == {"mydriver": update}

    d.pop_child("mydriver")
    assert d.children == {}

    with pytest.raises(AttributeError):
        d.pop_child("mydriver")


@pytest.mark.parametrize("child_name, values, expected", [
    (None, ['ab', 'ac', 'aa'], does_not_raise()),
    (None, ['aa', 'ac', 'ab'], does_not_raise()),
    (None, ('aa', 'ac', 'ab'), does_not_raise()),  # tuples work
    (None, {'aa', 'ac', 'ab'}, pytest.raises(TypeError,
        match="exec_order must be an ordered sequence")),  # sets fail
    (None, ['foo', 1, True], pytest.raises(ValueError,
        match="exec_order must be a permutation of \['aa', 'ab', 'ac'\]")),
    (None, ['aa', 'ac', 'aa', 'ab', 'ac', 'aa'],
        pytest.raises(ValueError, match="Repeated items \['aa', 'ac'\]")),
    ('aa', ['aab', 'aaa'], does_not_raise()),
    ('aa', ['aab', 'aaa', 'extra'], pytest.raises(ValueError)),  # too many elements
    ('aa', ['aab'], pytest.raises(ValueError)),  # too few elements
    ('ab', ['foo'], pytest.raises(ValueError, match="'ab' has no children")),
    ('ab', [], does_not_raise()),
])
def test_Module_exec_order(composite, child_name, values, expected):
    try:
        system = composite.children[child_name]
    except KeyError:
        system = composite

    with expected:
        system.exec_order = values
        assert isinstance(system.exec_order, MappingView)
        exec_order = list(system.exec_order)
        assert exec_order == list(values)
        assert exec_order == list(system.children)


@pytest.mark.parametrize("names", itertools.permutations(['aa', 'ab', 'ac']))
def test_Module_exec_order_perms(composite, names):
    composite.exec_order = names
    print(type(composite.exec_order))
    assert isinstance(composite.exec_order, MappingView)
    exec_order = list(composite.exec_order)
    assert exec_order == list(names)
    assert exec_order == list(composite.children)


def test_Module_setup_ran(fake):
    with mock.patch("cosapp.core.signal.signal.inspect"):
        fake_callback = mock.Mock(spec=lambda **kwargs: None)
        fake_callback.return_value = None

        fake.setup_ran.connect(Slot(fake_callback))
        fake.call_setup_run()

        fake_callback.assert_called_once_with()


def test_Module_computed(fake):
    with mock.patch("cosapp.core.signal.signal.inspect"):
        fake_callback = mock.Mock(spec=lambda **kwargs: None)
        fake_callback.return_value = None

        fake.computed.connect(Slot(fake_callback))
        fake.run_once()

        fake_callback.assert_called_once_with()


def test_Module_clean_ran(fake):
    with mock.patch("cosapp.core.signal.signal.inspect"):
        fake_callback = mock.Mock(spec=lambda **kwargs: None)
        fake_callback.return_value = None

        fake.clean_ran.connect(Slot(fake_callback))
        fake.call_clean_run()

        fake_callback.assert_called_once_with()


@pytest.mark.parametrize("format", LogFormat)
@pytest.mark.parametrize("msg, kwargs, to_log, emitted", [
    ("zombie call_setup_run", dict(), False, None),
    ("useless start call_clean_run", dict(activate=True), False, None),
    (f"{Module.CONTEXT_EXIT_MESSAGE} call_clean_run", dict(activate=False), False, dict(levelno=LogLevel.DEBUG, pattern=r"Compute calls for [\w\.]+: \d+")),
    ("common message", dict(), True, None),
])
def test_Module_log_debug_message(format, msg, kwargs, to_log, emitted):
    handler = mock.MagicMock(level=LogLevel.DEBUG, log=mock.MagicMock())
    rec = logging.getLogRecordFactory()("log_test", LogLevel.INFO, __file__, 22, msg, (), None)
    for key, value in kwargs.items():
        setattr(rec, key, value)
    
    m = Module("dummy")

    assert m.log_debug_message(handler, rec, format) == to_log

    if emitted:
        handler.log.assert_called_once()
        args = handler.log.call_args[0]
        assert args[0] == emitted["levelno"]
        assert re.match(emitted["pattern"], args[1]) is not None
    else:
        handler.log.assert_not_called()


def test_Module_tree_default(composite):
    names = [elem.name for elem in composite.tree()]
    assert names == [
        'aaa', 'aab', 'aa',
        'ab',
        'aca', 'acb', 'acc', 'ac',
        'a',
    ]


@pytest.mark.parametrize("downwards, expected", [
    (True,  ['a', 'aa', 'aaa', 'aab', 'ab', 'ac', 'aca', 'acb', 'acc']),
    (False, ['aaa', 'aab', 'aa', 'ab', 'aca', 'acb', 'acc', 'ac', 'a']),
])
def test_Module_tree(composite, downwards, expected):
    """Test top-to-bottom and bottom-to-top `Module.tree` generator"""
    names = [elem.name for elem in composite.tree(downwards)]
    assert names == expected


def test_Module_iter_tree(composite):
    a = composite
    elems = a.tree(downwards=True)
    assert next(elems) is a
    assert next(elems) is a.aa
    assert next(elems) is a.aa.aaa
    assert next(elems) is a.aa.aab
    assert next(elems) is a.ab
    # Check remaining elements
    assert [elem.name for elem in elems] == ['ac', 'aca', 'acb', 'acc']


@pytest.mark.parametrize("downwards, order, expected", [
    (True,  ['aa', 'ac', 'ab'], ['a', 'aa', 'aaa', 'aab', 'ac', 'aca', 'acb', 'acc', 'ab']),
    (True,  ['ac', 'ab', 'aa'], ['a', 'ac', 'aca', 'acb', 'acc', 'ab', 'aa', 'aaa', 'aab']),
    (False, ['aa', 'ac', 'ab'], ['aaa', 'aab', 'aa', 'aca', 'acb', 'acc', 'ac', 'ab', 'a']),
    (False, ['ac', 'ab', 'aa'], ['aca', 'acb', 'acc', 'ac', 'ab', 'aaa', 'aab', 'aa', 'a']),
])
def test_Module_tree_exec_order_top(composite, downwards, order, expected):
    """Check that `Module.tree` is consistent with `exec_order`"""
    composite.exec_order = order
    names = [elem.name for elem in composite.tree(downwards)]
    assert names == expected


def test_Module_tree_exec_order_sub(composite):
    """Check that `Module.tree` is consistent with `exec_order`"""
    a = composite
    names = lambda system, downwards=False: [s.name for s in system.tree(downwards)]

    a.exec_order = ('ac', 'ab', 'aa')
    assert names(a) == ['aca', 'acb', 'acc', 'ac', 'ab', 'aaa', 'aab', 'aa', 'a']

    a.aa.exec_order = ('aab', 'aaa')
    assert names(a) == ['aca', 'acb', 'acc', 'ac', 'ab', 'aab', 'aaa', 'aa', 'a']

    a.ac.exec_order = ('acb', 'acc', 'aca')
    assert names(a) == ['acb', 'acc', 'aca', 'ac', 'ab', 'aab', 'aaa', 'aa', 'a']
    assert names(a.aa) == ['aab', 'aaa', 'aa']
    assert names(a.ab) == ['ab']
    assert names(a.ac) == ['acb', 'acc', 'aca', 'ac']

    # Test downward iterator
    assert names(a, True) == ['a', 'ac', 'acb', 'acc', 'aca', 'ab', 'aa', 'aab', 'aaa']
    assert names(a.aa, True) == ['aa', 'aab', 'aaa']


def test_Module_tree_reverse(composite):
    """Check that changing `Module.tree` direction in a composite system
     does *not* produce a globally reversed iterator, owing to recursion.
    """
    downwards = list(composite.tree(downwards=True))
    upwards = list(composite.tree(downwards=False))

    assert upwards != list(reversed(downwards))



"""Mardown viewers for Module and Port."""
from cosapp.ports.port import BasePort
from cosapp.systems import System
from typing import List, Union


class PortMarkdownFormatter:
    """Markdown table formatter for ports"""
    
    def __init__(self, port: BasePort) -> None:
        self.port = port

    def content(self, contextual=True) -> List[str]:
        """Returns the port representation in Markdown format, as a list of strings.

        Parameters
        ----------
        contextual: bool
            If `True` (default), uses port full name; if `False`, displays only port name.

        Returns
        -------
        List[str]
            List of Markdown strings representing port variables as a table, with header.
        """
        port = self.port
        name = port.full_name() if contextual else port.name
        info = type(port).__name__
        if (desc := port.description):
            info += f", {desc}"
        doc = []
        doc.append(f"`{name}`: {info}")
        doc.extend(self.var_repr())
        return doc

    def var_repr(self) -> List[str]:
        """Returns the representation of port variables in Markdown format,
        as a list of strings (same as `content`, without name header).

        Returns
        -------
        List[str]
            List of Markdown strings representing port variables as a table.
        """
        content = [
            f"  {variable._repr_markdown_()}"
            for variable in self.port.variables()
        ]
        return self.wrap(content)

    def markdown(self, contextual=True) -> str:
        """Returns the port representation in Markdown format.

        Parameters
        ----------
        contextual: bool
            If `True` (default), uses port full name; if `False`, displays only port name.

        Returns
        -------
        str
            Markdown formatted representation
        """
        return "\n".join(self.content(contextual))

    @classmethod
    def wrap(cls, content: Union[str, List[str]]) -> List[str]:
        if isinstance(content, str):
            content = [content]
        header = ["|  |  |", "---|---"]
        doc = ["", cls.div_header(), ""]
        doc.extend(header)
        doc.extend(content)
        doc.append("</div>")
        return doc

    @classmethod
    def div_header(cls) -> str:
        """Div header used to override Jupyter Lab table CSS"""
        return "".join([
            r"<div class='cosapp-port-table' style='margin-left: 25px; margin-top: -12px'>",
            r"<style type='text/css'>",
            r".cosapp-port-table >table >thead{display: none}",  # suppress empty table header
            r".cosapp-port-table tbody tr{background: rgba(255, 255, 255, 0)!important}",  # override even/odd coloring
            r".cosapp-port-table tbody tr:hover{background: #e1f5fe!important}",  # set hover color
            r".cosapp-port-table >table {margin-left: unset; margin-right: unset}",
            r"</style>",
        ])


def port_to_md(port: BasePort) -> str:
    """Returns the representation of `port` in Markdown format.

    Parameters
    ----------
    port: BasePort
        Port to describe

    Returns
    -------
    str
        Markdown formatted representation
    """
    formatter = PortMarkdownFormatter(port)
    return formatter.markdown()


def system_to_md(system: System) -> str:
    """Returns the representation of `system` in Markdown format.

    Parameters
    ----------
    system: System
        System to describe

    Returns
    -------
    str
        Markdown formatted representation
    """
    doc = list()
    if system.tags:
        doc.extend(["", f"**Tags**: {list(system.tags)!s}", ""])

    def get_child_doc(s: System) -> str:
        info = type(s).__name__
        if (desc := s.description):
            info += f", {desc}"
        return f"- `{s.name}`: {info}"

    if system.children:
        doc.extend(["", "### Child components", ""])
        doc.extend(map(get_child_doc, system.children.values()))

    def dump_port_data(header, port_dict):
        port_docs = []
        for port in port_dict.values():
            if len(port) > 0:
                formatter = PortMarkdownFormatter(port)
                port_doc = formatter.content(contextual=False)
                port_doc[0] = f"- {port_doc[0]}"
                port_docs.extend(port_doc)
        if port_docs:
            doc.extend(["", f"### {header.title()}", ""])
            doc.extend(port_docs)

    dump_port_data("inputs", system.inputs)
    dump_port_data("outputs", system.outputs)

    if system.residues:
        doc.extend(["", "### Residues", ""])
        doc.append(", ".join(f"`{key}`" for key in system.residues))

    return "\n".join(doc)



import numpy
from typing import Any, Iterable, Dict, Optional, Union, List

from cosapp.core.eval_str import AssignString
from cosapp.core.numerics.basics import MathematicalProblem
from cosapp.core.numerics.boundary import Boundary
from cosapp.drivers.iterativecase import IterativeCase
from cosapp.drivers.utils import DesignProblemHandler
from cosapp.systems import System
from cosapp.utils.helpers import check_arg

import logging
logger = logging.getLogger(__name__)


class RunSingleCase(IterativeCase):
    """Set new boundary conditions and equations on the system.

    By default, it has a :py:class:`~cosapp.drivers.runonce.RunOnce` driver as child to run the system.

    Attributes
    ----------
    case_values : List[AssignString]
        List of requested variable assignments to set up the case
    initial_values : Dict[str, Tuple[Any, Optional[numpy.ndarray]]]
        List of variables to set with the values to set and associated indices selection

    Parameters
    ----------
    name : str
        Name of the driver
    owner : System, optional
        :py:class:`~cosapp.systems.system.System` to which driver belongs; defaults to `None`
    **kwargs : Any
        Keyword arguments will be used to set driver options
    """

    __slots__ = ('__case_values', '__raw_problem', '__processed', 'problem')

    def __init__(
        self,
        name: str,
        owner: Optional[System] = None,
        **kwargs
        ) -> None:
        """Initialize driver

        Parameters
        ----------
        name: str, optional
            Name of the `Module`
        owner: System, optional
            :py:class:`~cosapp.systems.system.System` to which driver belongs; defaults to `None`
        **kwargs : Dict[str, Any]
            Optional keywords arguments formwarded to base class.
        """
        super().__init__(name, owner, **kwargs)
        self.__case_values = []  # type: List[AssignString]
            # desc="List of assignments 'lhs <- rhs' to perform in the present case.")
        self.problem = None # type: Optional[MathematicalProblem]
            # desc='Full mathematical problem to be solved on this case.'
        self.__raw_problem = DesignProblemHandler(owner)
        self.__processed = DesignProblemHandler(owner)
        self.owner = owner

    @property
    def design(self) -> MathematicalProblem:
        """MathematicalProblem: Design problem solved for case"""
        return self.__raw_problem.design

    @property
    def offdesign(self) -> MathematicalProblem:
        """MathematicalProblem: Local problem solved for case"""
        return self.__raw_problem.offdesign

    @property
    def processed_problems(self) -> DesignProblemHandler:
        """DesignProblemHandler: design/off-design problem handler"""
        return self.__processed

    def reset_problem(self) -> None:
        """Reset design and off-design problems defined on case."""
        self.__raw_problem = DesignProblemHandler(self.owner)
        self.__processed = DesignProblemHandler(self.owner)
        self.problem = None

    def __merge_problems(self) -> None:
        self.__activate_targets()
        self.problem = self.merged_problem(copy=False)

    def __activate_targets(self) -> None:
        """Activate targets in processed problems"""
        extract_targets = lambda problem: set(
            r.target for r in problem.deferred_residues.values()
        )
        target_names = set.union(
            *map(extract_targets, self.__processed.problems)
        )
        if target_names:
            # Set init values corresponding to targetted variables
            for boundary in self.initial_values.values():
                if boundary.basename in target_names:
                    boundary.set_to_default()
            for problem in self.__processed.problems:
                problem.activate_targets()

    def merged_problem(self, copy=True) -> MathematicalProblem:
        handler = self.__processed
        name = self.name
        try:
            return handler.merged_problem(name=name, offdesign_prefix=None, copy=copy)
        except ValueError as error:
            error.args = (f"{error.args[0]} in {name!r}",)
            raise

    def setup_run(self):
        """Method called once before starting any simulation."""
        super().setup_run()
        
        raw = self.__raw_problem
        # Transfer problem copies from `raw` to `processed`
        processed = raw.copy(prune=False)

        # Add owner off-design problem to `processed.offdesign`
        owner_problem = self.owner.assembled_problem()
        processed.offdesign.extend(owner_problem)

        # Resolve unknown aliasing in `processed`
        self.__processed = processed.copy(prune=True)
        self.__merge_problems()

    def add_offdesign_problem(self, offdesign: MathematicalProblem) -> MathematicalProblem:
        """Add outer off-design problem to inner problem.

        Returns:
        ----------
        - `MathematicalProblem`
            The modified mathematical problem
        """
        # Unknowns & residues are duplicated to avoid side effects between points
        # Existing unknowns and equations are silently overwritten.
        if not offdesign.is_empty():
            self.__processed.offdesign.extend(offdesign, copy=True, overwrite=True)
            self.__merge_problems()
        return self.problem

    def _precompute(self) -> None:
        """Actions to carry out before the :py:meth:`~cosapp.drivers.runonce.RunOnce.compute` method call.

        It sets the boundary conditions and changes variable status.
        """
        super()._precompute()

        # Set boundary conditions
        self.apply_values()

        # Set offdesign variables
        design_unknowns = set(self.design.unknowns)
        problem = self.get_problem()
        for name, unknown in problem.unknowns.items():
            if name in design_unknowns:
                continue
            if not numpy.array_equal(unknown.value, unknown.default_value):
                unknown.set_to_default()

    def apply_values(self) -> None:
        owner_changed = False
        for assignment in self.case_values:
            value, changed = assignment.exec()
            if changed:
                owner_changed = True

        if owner_changed:
            self.owner.touch()

    def clean_run(self):
        """Method called once after any simulation."""
        self.problem = None

    def get_problem(self) -> MathematicalProblem:
        """Returns the full mathematical for the case.

        Returns
        -------
        MathematicalProblem
            The full mathematical problem to solve for the case
        """
        if self.problem is None:
            logger.warning("RunSingleCase.get_problem called with no prior call to RunSingleCase.setup_run.")
            return MathematicalProblem(self.name, self.owner)
        else:
            return self.problem

    def set_values(self, modifications: Dict[str, Any]) -> None:
        """Enter the set of variables defining the case, from a dictionary of the kind {'variable1': value1, ...}
        Note: will erase all previously defined values. Use 'add_values' to append new case values.

        The variable can be contextual `child1.port2.var`. The only rule is that it should belong to
        the owner `System` of this driver or any of its descendants.

        Parameters
        ----------
        modifications : Dict[str, Any]
            Dictionary of (variable name, value)

        Examples
        --------
        >>> driver.set_values({'myvar': 42, 'port.dummy': 'banana'})
        """
        self.clear_values()
        self.add_values(modifications)

    def add_values(self, modifications: Dict[str, Any]) -> None:
        """Add a set of variables to the list of case values, from a dictionary of the kind {'variable1': value1, ...}

        The variable can be contextual `child1.port2.var`. The only rule is that it should belong to
        the owner `System` of this driver or any of its descendants.

        Parameters
        ----------
        modifications : Dict[str, Any]
            Dictionary of (variable name, value)

        Examples
        --------
        >>> driver.add_values({'myvar': 42, 'port.dummy': 'banana'})
        """
        owner = self.owner
        if owner is None:
            raise AttributeError(
                f"Driver {self.name!r} must be attached to a System to set case values."
            )
        check_arg(modifications, 'modifications', dict)

        init = {}
        for varname, value in modifications.items():
            info = Boundary.parse(owner, varname)  # checks that variable is valid
            if info.port.is_input:
                self.__case_values.append(AssignString(varname, value, owner))
            else:
                init[varname] = value
        if init:
            varnames = list(init)
            if len(init) == 1:
                head = f"Variable {varnames[0]}"
                tail = f"has been set as initial condition"
            else:
                head = f"Variables {varnames}"
                tail = f"have been set as initial conditions"
            logger.info(
                f"{head} declared in `{self.name}` values {tail}."
            )
            self.set_init(init)

    def add_value(self, variable: str, value: Any) -> None:
        """Add a single variable to list of case values.

        The variable can be contextual `child1.port2.var`. The only rule is that it should belong to
        the owner `System` of this driver or any of its descendants.

        Parameters
        ----------
        variable : str
            Name of the variable
        value : Any
            Value to be used.

        Examples
        --------
        >>> driver.add_value('myvar', 42)
        """
        self.add_values({variable: value})

    def clear_values(self):
        self.__case_values.clear()

    @property
    def case_values(self) -> List[AssignString]:
        return self.__case_values

    def extend(self, problem: MathematicalProblem) -> MathematicalProblem:
        """Extend local problem. Shortcut to `self.offdesign.extend(problem)`.
        
        Parameters
        ----------
        - problem: MathematicalProblem

        Returns
        -------
        MathematicalProblem
            The extended mathematical problem
        """
        return self.offdesign.extend(problem)

    def add_unknown(self,
        name: Union[str, Iterable[Union[dict, str]]],
        *args, **kwargs,
    ) -> MathematicalProblem:
        """Add local unknown(s).
        Shortcut to `self.offdesign.add_unknown(name, *args, **kwargs)`.

        More details in `MathematicalProblem.add_unknown`.

        Parameters
        ----------
        - name: str or Iterable of dictionary or str
            Name of the variable or list of variables to add
        - *args, **kwargs: Forwarded to `MathematicalProblem.add_unknown`

        Returns
        -------
        MathematicalProblem
            The modified mathematical problem
        """
        return self.offdesign.add_unknown(name, *args, **kwargs)

    def add_equation(self,
        equation: Union[str, Iterable[Union[dict, str]]],
        *args, **kwargs,
    ) -> MathematicalProblem:
        """Add local equation(s).
        Shortcut to `self.offdesign.add_equation(equation, *args, **kwargs)`.

        More details in `MathematicalProblem.add_equation`.

        Parameters
        ----------
        - equation: str or Iterable of str of the kind 'lhs == rhs'
            Equation or list of equations to add
        - *args, **kwargs: Forwarded to `MathematicalProblem.add_equation`

        Returns
        -------
        MathematicalProblem
            The modified mathematical problem
        """
        return self.offdesign.add_equation(equation, *args, **kwargs)

    def add_target(self,
        expression: Union[str, Iterable[str]],
        *args, **kwargs,
    ) -> MathematicalProblem:
        """Add deferred equation(s) on current point.
        Shortcut to `self.offdesign.add_target(expression, *args, **kwargs)`.

        More details in `MathematicalProblem.add_target`.

        Parameters
        ----------
        - expression: str
            Targetted expression
        - *args, **kwargs : Forwarded to `MathematicalProblem.add_target`

        Returns
        -------
        MathematicalProblem
            The modified mathematical problem
        """
        return self.offdesign.add_target(expression, *args, **kwargs)



import abc
from typing import List


class Component(metaclass=abc.ABCMeta):
    """Abstract Base Class for visited components"""
    __slots__ = ()

    @abc.abstractmethod
    def accept(self, visitor) -> None:
        """Determines class-dependent course of action when visited by `visitor`"""
        pass


class Visitor:
    """Base class for visitors"""
    def visit_system(self, system) -> None:
        pass

    def visit_port(self, port) -> None:
        pass

    def visit_driver(self, driver) -> None:
        pass


def send(visitor: Visitor, components: List[Component]):
    """Send a visitor to a list of generic components"""
    for component in components:
        component.accept(visitor)



""" Unit tests for the problem interface."""

import pytest
import unittest
import os

from tempfile import mkdtemp
from collections import OrderedDict

from cosapp.base import System, Port
from cosapp.tools.problem_viewer.problem_viewer import _get_viewer_data, view_model
from cosapp.tests.library.systems.others import ComplexTurbofan


def dict_to_tuples(d: dict):
    return tuple(map(tuple, d.items()))


class TestViewModelData(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None
        self.dir = mkdtemp()
        self.problem_filename = os.path.join(self.dir, "problem_n2")
        self.problem_html_filename = self.problem_filename
        self.expected_tree = OrderedDict(
            [
                ("name", "turbofan"),
                ("type", "subsystem"),
                ("subsystem_type", "group"),
                (
                    "children",
                    [
                        OrderedDict(
                            [
                                ("name", "atm"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "inlet"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "W_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        )
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "fanC"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "gh"),
                                                                ("type", "param"),
                                                            ]
                                                        )
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "mech_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "XN"),                                                                
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "PW"),                                                                
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "ductC"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "group"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "fl_in"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Tt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Pt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "W",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "merger"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "group",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl1_in",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl2_in",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl_out",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "duct"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "group",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "inwards",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "A",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "cst_loss",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "glp",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl_in",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "outwards",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "PR",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        )
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl_out",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "bleed"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "group",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "inwards",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "split_ratio",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        )
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl_in",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl1_out",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl2_out",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "fl_out"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Tt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Pt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "W",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fan"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "group"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "inwards"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "gh",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        )
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "fl_in"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Tt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Pt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "W",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "mech_in"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "XN",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "PW",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "outwards"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "pcnr",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "pr",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "effis",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "wr",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "PWfan",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "fl_out"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Tt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Pt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "W",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "merger"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "fl1_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl2_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "duct"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "A"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "cst_loss"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "glp"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "outwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "PR"),
                                                                ("type", "unknown"),
                                                            ]
                                                        )
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "bleed"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "split_ratio"),
                                                                ("type", "param"),
                                                            ]
                                                        )
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl1_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl2_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "noz"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Acol"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Aexit"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "outwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "WRnozzle"),
                                                                ("type", "unknown"),
                                                            ]
                                                        )
                                                    ],
                                                ),

                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                    ],
                ),
            ]
        )

        self.expected_conns = [
            dict(src="atm.fl_out.Tt", tgt="inlet.fl_in.Tt"),
            dict(src="atm.fl_out.Pt", tgt="inlet.fl_in.Pt"),
            dict(src="inlet.fl_out.Tt", tgt="fanC.fl_in.Tt"),
            dict(src="inlet.fl_out.Pt", tgt="fanC.fl_in.Pt"),
            dict(src="inlet.fl_out.W", tgt="fanC.fl_in.W"),
            dict(src="fanC.fl_out.Tt", tgt="merger.fl1_in.Tt"),
            dict(src="fanC.fl_out.Pt", tgt="merger.fl1_in.Pt"),
            dict(src="fanC.fl_out.W", tgt="merger.fl1_in.W"),
            dict(src="bleed.fl2_out.Tt", tgt="merger.fl2_in.Tt"),
            dict(src="bleed.fl2_out.Pt", tgt="merger.fl2_in.Pt"),
            dict(src="bleed.fl2_out.W", tgt="merger.fl2_in.W"),
            dict(src="merger.fl_out.Tt", tgt="duct.fl_in.Tt"),
            dict(src="merger.fl_out.Pt", tgt="duct.fl_in.Pt"),
            dict(src="merger.fl_out.W", tgt="duct.fl_in.W"),
            dict(src="duct.fl_out.Tt", tgt="bleed.fl_in.Tt"),
            dict(src="duct.fl_out.Pt", tgt="bleed.fl_in.Pt"),
            dict(src="duct.fl_out.W", tgt="bleed.fl_in.W"),
            dict(src="bleed.fl1_out.Tt", tgt="noz.fl_in.Tt"),
            dict(src="bleed.fl1_out.Pt", tgt="noz.fl_in.Pt"),
            dict(src="bleed.fl1_out.W", tgt="noz.fl_in.W"),
            dict(src="fanC.fl_in.Tt", tgt="fanC.ductC.fl_in.Tt"),
            dict(src="fanC.fl_in.Pt", tgt="fanC.ductC.fl_in.Pt"),
            dict(src="fanC.fl_in.W", tgt="fanC.ductC.fl_in.W"),
            dict(src="fanC.mech_in.XN", tgt="fanC.fan.mech_in.XN"),
            dict(src="fanC.mech_in.PW", tgt="fanC.fan.mech_in.PW"),
            dict(src="fanC.inwards.gh", tgt="fanC.fan.inwards.gh"),
            dict(src="fanC.ductC.fl_out.Tt", tgt="fanC.fan.fl_in.Tt"),
            dict(src="fanC.ductC.fl_out.Pt", tgt="fanC.fan.fl_in.Pt"),
            dict(src="fanC.ductC.fl_out.W", tgt="fanC.fan.fl_in.W"),
            dict(src="fanC.fan.fl_out.Tt", tgt="fanC.fl_out.Tt"),
            dict(src="fanC.fan.fl_out.Pt", tgt="fanC.fl_out.Pt"),
            dict(src="fanC.fan.fl_out.W", tgt="fanC.fl_out.W"),
            dict(src="fanC.ductC.fl_in.Tt", tgt="fanC.ductC.merger.fl1_in.Tt"),
            dict(src="fanC.ductC.fl_in.Pt", tgt="fanC.ductC.merger.fl1_in.Pt"),
            dict(src="fanC.ductC.fl_in.W", tgt="fanC.ductC.merger.fl1_in.W"),
            dict(src="fanC.ductC.bleed.fl2_out.Tt", tgt="fanC.ductC.merger.fl2_in.Tt"),
            dict(src="fanC.ductC.bleed.fl2_out.Pt", tgt="fanC.ductC.merger.fl2_in.Pt"),
            dict(src="fanC.ductC.bleed.fl2_out.W", tgt="fanC.ductC.merger.fl2_in.W"),
            dict(src="fanC.ductC.bleed.fl1_out.Tt", tgt="fanC.ductC.fl_out.Tt"),
            dict(src="fanC.ductC.bleed.fl1_out.Pt", tgt="fanC.ductC.fl_out.Pt"),
            dict(src="fanC.ductC.bleed.fl1_out.W", tgt="fanC.ductC.fl_out.W"),
            dict(src="fanC.ductC.merger.fl_out.Tt", tgt="fanC.ductC.duct.fl_in.Tt"),
            dict(src="fanC.ductC.merger.fl_out.Pt", tgt="fanC.ductC.duct.fl_in.Pt"),
            dict(src="fanC.ductC.merger.fl_out.W", tgt="fanC.ductC.duct.fl_in.W"),
            dict(src="fanC.ductC.duct.fl_out.Tt", tgt="fanC.ductC.bleed.fl_in.Tt"),
            dict(src="fanC.ductC.duct.fl_out.Pt", tgt="fanC.ductC.bleed.fl_in.Pt"),
            dict(src="fanC.ductC.duct.fl_out.W", tgt="fanC.ductC.bleed.fl_in.W"),
        ]

        self.expected_tree_default = OrderedDict(
            [
                ("name", "turbofan"),
                ("type", "subsystem"),
                ("subsystem_type", "group"),
                (
                    "children",
                    [
                        OrderedDict(
                            [
                                ("name", "atm"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                ("children", []),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "inlet"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "W_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        )
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "fanC"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                ("children", []),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "mech_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "XN"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "PW"),                                                                
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "ductC"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "group"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "fl_in"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Tt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Pt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "W",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "merger"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "group",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl1_in",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl2_in",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl_out",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "duct"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "group",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "inwards",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl_in",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "outwards",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl_out",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "bleed"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "group",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "inwards",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl_in",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl1_out",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl2_out",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "fl_out"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Tt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Pt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "W",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fan"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "group"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "inwards"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                ("children", []),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "fl_in"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Tt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Pt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "W",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "mech_in"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "XN",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "PW",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "outwards"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                ("children", []),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "fl_out"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Tt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Pt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "W",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "merger"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "fl1_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl2_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "duct"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                ("children", []),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "outwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                ("children", []),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "bleed"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                ("children", []),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl1_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl2_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "noz"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                ("children", []),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "outwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                ("children", []),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                    ],
                ),
            ]
        )

        self.expected_conns_default = [
            dict(src="atm.fl_out.Tt", tgt="inlet.fl_in.Tt"),
            dict(src="atm.fl_out.Pt", tgt="inlet.fl_in.Pt"),
            dict(src="inlet.fl_out.Tt", tgt="fanC.fl_in.Tt"),
            dict(src="inlet.fl_out.Pt", tgt="fanC.fl_in.Pt"),
            dict(src="inlet.fl_out.W", tgt="fanC.fl_in.W"),
            dict(src="fanC.fl_out.Tt", tgt="merger.fl1_in.Tt"),
            dict(src="fanC.fl_out.Pt", tgt="merger.fl1_in.Pt"),
            dict(src="fanC.fl_out.W", tgt="merger.fl1_in.W"),
            dict(src="bleed.fl2_out.Tt", tgt="merger.fl2_in.Tt"),
            dict(src="bleed.fl2_out.Pt", tgt="merger.fl2_in.Pt"),
            dict(src="bleed.fl2_out.W", tgt="merger.fl2_in.W"),
            dict(src="merger.fl_out.Tt", tgt="duct.fl_in.Tt"),
            dict(src="merger.fl_out.Pt", tgt="duct.fl_in.Pt"),
            dict(src="merger.fl_out.W", tgt="duct.fl_in.W"),
            dict(src="duct.fl_out.Tt", tgt="bleed.fl_in.Tt"),
            dict(src="duct.fl_out.Pt", tgt="bleed.fl_in.Pt"),
            dict(src="duct.fl_out.W", tgt="bleed.fl_in.W"),
            dict(src="bleed.fl1_out.Tt", tgt="noz.fl_in.Tt"),
            dict(src="bleed.fl1_out.Pt", tgt="noz.fl_in.Pt"),
            dict(src="bleed.fl1_out.W", tgt="noz.fl_in.W"),
            dict(src="fanC.fl_in.Tt", tgt="fanC.ductC.fl_in.Tt"),
            dict(src="fanC.fl_in.Pt", tgt="fanC.ductC.fl_in.Pt"),
            dict(src="fanC.fl_in.W", tgt="fanC.ductC.fl_in.W"),
            dict(src="fanC.mech_in.XN", tgt="fanC.fan.mech_in.XN"),
            dict(src="fanC.mech_in.PW", tgt="fanC.fan.mech_in.PW"),
            dict(src="fanC.ductC.fl_out.Tt", tgt="fanC.fan.fl_in.Tt"),
            dict(src="fanC.ductC.fl_out.Pt", tgt="fanC.fan.fl_in.Pt"),
            dict(src="fanC.ductC.fl_out.W", tgt="fanC.fan.fl_in.W"),
            dict(src="fanC.fan.fl_out.Tt", tgt="fanC.fl_out.Tt"),
            dict(src="fanC.fan.fl_out.Pt", tgt="fanC.fl_out.Pt"),
            dict(src="fanC.fan.fl_out.W", tgt="fanC.fl_out.W"),
            dict(src="fanC.ductC.fl_in.Tt", tgt="fanC.ductC.merger.fl1_in.Tt"),
            dict(src="fanC.ductC.fl_in.Pt", tgt="fanC.ductC.merger.fl1_in.Pt"),
            dict(src="fanC.ductC.fl_in.W", tgt="fanC.ductC.merger.fl1_in.W"),
            dict(src="fanC.ductC.bleed.fl2_out.Tt", tgt="fanC.ductC.merger.fl2_in.Tt"),
            dict(src="fanC.ductC.bleed.fl2_out.Pt", tgt="fanC.ductC.merger.fl2_in.Pt"),
            dict(src="fanC.ductC.bleed.fl2_out.W", tgt="fanC.ductC.merger.fl2_in.W"),
            dict(src="fanC.ductC.bleed.fl1_out.Tt", tgt="fanC.ductC.fl_out.Tt"),
            dict(src="fanC.ductC.bleed.fl1_out.Pt", tgt="fanC.ductC.fl_out.Pt"),
            dict(src="fanC.ductC.bleed.fl1_out.W", tgt="fanC.ductC.fl_out.W"),
            dict(src="fanC.ductC.merger.fl_out.Tt", tgt="fanC.ductC.duct.fl_in.Tt"),
            dict(src="fanC.ductC.merger.fl_out.Pt", tgt="fanC.ductC.duct.fl_in.Pt"),
            dict(src="fanC.ductC.merger.fl_out.W", tgt="fanC.ductC.duct.fl_in.W"),
            dict(src="fanC.ductC.duct.fl_out.Tt", tgt="fanC.ductC.bleed.fl_in.Tt"),
            dict(src="fanC.ductC.duct.fl_out.Pt", tgt="fanC.ductC.bleed.fl_in.Pt"),
            dict(src="fanC.ductC.duct.fl_out.W", tgt="fanC.ductC.bleed.fl_in.W"),
        ]

    @staticmethod
    def make_system():
        return ComplexTurbofan("turbofan")

    def test_model_viewer_has_correct_data_from_problem(self):
        """
        Verify that the correct model structure data exists when stored as compared
        to the expected structure, using the SellarStateConnection model.
        """
        p = self.make_system()

        model_viewer_data = _get_viewer_data(p, include_orphan_vars=True)

        assert model_viewer_data["tree"] == self.expected_tree
        # Check connection list, regardless of order
        actual = set(map(dict_to_tuples, model_viewer_data["connections_list"]))
        expected = set(map(dict_to_tuples, self.expected_conns))
        assert actual == expected

    def test_model_viewer_takes_into_account_include_wards(self):
        """
        Test that model viewer does not show orphan vars by default.
        """
        p = self.make_system()

        model_viewer_data = _get_viewer_data(p, include_orphan_vars=False)
        # tree_json = model_viewer_data["tree"]
        # conns_json = model_viewer_data["connections_list"]

        assert model_viewer_data["tree"] == self.expected_tree_default
        # Check connection list, regardless of order
        actual = set(map(dict_to_tuples, model_viewer_data["connections_list"]))
        expected = set(map(dict_to_tuples, self.expected_conns_default))
        assert actual == expected

    def test_view_model_from_problem(self):
        """
        Test that an n2 html file is generated from a Problem.
        """
        p = self.make_system()

        view_model(p, outfile=self.problem_filename, show_browser=False)

        # Check that the html file has been created and has something in it.
        assert os.path.isfile(self.problem_html_filename),  f"{self.problem_html_filename} is not a valid file."
        assert os.path.getsize(self.problem_html_filename) > 100


def test_get_viewer_data_same_name():
    """Check that no confusion exists in connectors
    when a subsystem and its parent have the same name.
    """
    class XyPort(Port):
        def setup(self):
            self.add_variable("x", 1.0)
            self.add_variable("y", 1.0)

    class Bogus(System):
        def setup(self):
            self.add_input(XyPort, 'p_in')
            self.add_output(XyPort, 'p_out')
    
    top = System("top")
    a = Bogus("top")  # same name as parent
    b = Bogus("b")
    top.add_child(a)
    top.add_child(b, pulling='p_out')  # connector to `top.fl_out`
    top.connect(a.p_in, b.p_out)  # connector to `top.top.fl_out`
    top.exec_order = ['top', 'b']

    assert set(top.child_connectors) == {'top'}
    assert set(top.connectors()) == {
        'b.p_out -> p_out',
        'b.p_out -> top.p_in',
    }

    data = _get_viewer_data(top, True)
    expected_connections = [
        {'src': 'b.p_out.x', 'tgt': 'p_out.x'},
        {'src': 'b.p_out.y', 'tgt': 'p_out.y'},
        {'src': 'b.p_out.x', 'tgt': 'top.p_in.x'},
        {'src': 'b.p_out.y', 'tgt': 'top.p_in.y'},
    ]
    # Check connection list, regardless of order
    actual_pairs = set(map(dict_to_tuples, data["connections_list"]))
    expected_pairs = set(map(dict_to_tuples, expected_connections))
    assert actual_pairs == expected_pairs
